# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Marshmallow fields for search parameter deserialization."""

from invenio_i18n import lazy_gettext as _
from marshmallow import fields


class ReferenceString(fields.Field):
    """Field for serializing reference querystring parameters into reference dicts.

    Reference querystring parameters are of the shape ``"TYPE:ID"``, which are
    deserialized into reference dicts of the shape ``{"TYPE": "ID"}``.
    """

    #: Default error messages.
    default_error_messages = {
        "invalid": _("Not a valid reference."),
    }

    def _deserialize(self, value, attr, data, **kwargs):
        """Deserialize reference string into a dict."""
        if value is None:
            return None
        elif ":" not in value:
            raise self.make_error("invalid")

        # if there's multiple colons, we assume the first one is the separator
        type_, id_ = value.split(":", 1)
        return {type_: id_}

    def _serialize(self, value, attr, obj, **kwargs):
        """Serialize the reference dict to a reference string."""
        # TODO do we even need this?
        if not isinstance(value, dict) or len(value) != 1:
            # a reference dict needs to have exactly one key
            raise self.make_error("invalid")

        key, value = list(value.items())[0]
        return f"{key}:{value}"



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
#
# Invenio-Requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Request views decorators module."""

from functools import wraps

from flask import g

from invenio_requests.proxies import current_requests


def pass_request(expand=False):
    """Fetch the request record and pass it as kwarg."""

    def decorator(f):
        @wraps(f)
        def view(**kwargs):
            """Decorated view."""
            pid_value = kwargs["request_pid_value"]
            request = current_requests.requests_service.read(
                id_=pid_value, identity=g.identity, expand=expand
            )
            kwargs["request"] = request
            return f(**kwargs)

        return view

    return decorator



#
# This file is part of Invenio.
# Copyright (C) 2025 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Add index to `request_events.request_id` FK column."""

from alembic import op

# revision identifiers, used by Alembic.
revision = "1759321170"
down_revision = "a14fa442680f"
branch_labels = ()
depends_on = None


def upgrade():
    """Upgrade database."""
    op.create_index(
        op.f("ix_request_events_request_id"),
        "request_events",
        ["request_id"],
        unique=False,
    )


def downgrade():
    """Downgrade database."""
    op.drop_index(op.f("ix_request_events_request_id"), table_name="request_events")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 Graz University of Technology.
#
# Invenio-Requests is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Notification builders."""

from invenio_notifications.models import Notification
from invenio_notifications.registry import EntityResolverRegistry
from invenio_notifications.services.builders import NotificationBuilder
from invenio_notifications.services.filters import KeyRecipientFilter
from invenio_notifications.services.generators import EntityResolve, UserEmailBackend
from invenio_users_resources.notifications.filters import UserPreferencesRecipientFilter
from invenio_users_resources.notifications.generators import (
    EmailRecipient,
    IfEmailRecipient,
)

from invenio_requests.notifications.filters import UserRecipientFilter

from .generators import (
    CommentRepliesParticipantsRecipient,
    RequestParticipantsRecipient,
)


class CommentRequestEventCreateNotificationBuilder(NotificationBuilder):
    """Notification builder for comment request event creation."""

    type = "comment-request-event.create"

    @classmethod
    def build(cls, request, request_event):
        """Build notification with context."""
        return Notification(
            type=cls.type,
            context={
                "request": EntityResolverRegistry.reference_entity(request),
                "request_event": EntityResolverRegistry.reference_entity(request_event),
            },
        )

    context = [
        EntityResolve(key="request"),
        EntityResolve(key="request.created_by"),
        EntityResolve(key="request.receiver"),
        EntityResolve(key="request_event"),
        EntityResolve(key="request_event.created_by"),
    ]

    recipients = [
        RequestParticipantsRecipient(key="request"),
        IfEmailRecipient(
            key="request.created_by",
            then_=[EmailRecipient(key="request.created_by")],
            else_=[],
        ),
    ]

    recipient_filters = [
        # remove a possible email recipient
        KeyRecipientFilter(key="request_event.created_by"),
        # do not send notification to user creating the comment
        UserRecipientFilter(key="request_event.created_by"),
        UserPreferencesRecipientFilter(),
    ]

    recipient_backends = [
        UserEmailBackend(),
    ]


class CommentRequestEventReplyNotificationBuilder(
    CommentRequestEventCreateNotificationBuilder
):
    """Notification builder for reply creation.

    Notifies only participants in the specific comment thread
    (parent author and other reply authors), not all request participants.
    """

    type = "comment-request-event.reply"

    recipients = [
        CommentRepliesParticipantsRecipient(event_key="request_event", key="request"),
    ]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2022 CERN.
# Copyright (C) 2021-2022 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Requests resource."""


from flask import g
from flask_resources import resource_requestctx, response_handler, route
from invenio_records_resources.resources import RecordResource
from invenio_records_resources.resources.records.resource import (
    request_data,
    request_extra_args,
    request_headers,
    request_search_args,
    request_view_args,
)
from invenio_records_resources.resources.records.utils import search_preference


#
# Resource
#
class RequestsResource(RecordResource):
    """Resource for generic requests."""

    def create_blueprint(self, **options):
        """Create the blueprint."""
        # We avoid passing url_prefix to the blueprint because we need to
        # install URLs under both /records and /user/records. Instead we
        # add the prefix manually to each route (which is anyway what Flask
        # does in the end)
        options["url_prefix"] = ""
        return super().create_blueprint(**options)

    def create_url_rules(self):
        """Create the URL rules for the record resource."""
        routes = self.config.routes

        def p(route):
            """Prefix a route with the URL prefix."""
            return f"{self.config.url_prefix}{route}"

        def s(route):
            """Suffix a route with the URL prefix."""
            return f"{route}{self.config.url_prefix}"

        return [
            route("GET", p(routes["list"]), self.search),
            route("GET", p(routes["item"]), self.read),
            route("PUT", p(routes["item"]), self.update),
            route("DELETE", p(routes["item"]), self.delete),
            route("POST", p(routes["action"]), self.execute_action),
            route("GET", s(routes["user-prefix"]), self.search_user_requests),
            route("GET", p(routes["lock"]), self.lock_request),
            route("GET", p(routes["unlock"]), self.unlock_request),
        ]

    @request_extra_args
    @request_search_args
    @request_view_args
    @response_handler(many=True)
    def search(self):
        """Perform a search over the items."""
        hits = self.service.search(
            identity=g.identity,
            params=resource_requestctx.args,
            search_preference=search_preference(),
            expand=resource_requestctx.args.get("expand", False),
        )
        return hits.to_dict(), 200

    @request_extra_args
    @request_search_args
    @request_view_args
    @response_handler(many=True)
    def search_user_requests(self):
        """Perform a search over user requets.

        /GET /user/requests
        """
        hits = self.service.search_user_requests(
            identity=g.identity,
            params=resource_requestctx.args,
            search_preference=search_preference(),
            expand=resource_requestctx.args.get("expand", False),
        )
        return hits.to_dict(), 200

    @request_extra_args
    @request_view_args
    @response_handler()
    def read(self):
        """Read an item."""
        item = self.service.read(
            id_=resource_requestctx.view_args["id"],
            identity=g.identity,
            expand=resource_requestctx.args.get("expand", False),
        )
        return item.to_dict(), 200

    @request_extra_args
    @request_headers
    @request_view_args
    @request_data
    @response_handler()
    def update(self):
        """Update an item."""
        # TODO should we allow updating of requests in this general resource?
        item = self.service.update(
            id_=resource_requestctx.view_args["id"],
            identity=g.identity,
            data=resource_requestctx.data,
            expand=resource_requestctx.args.get("expand", False),
        )
        return item.to_dict(), 200

    @request_headers
    @request_view_args
    def delete(self):
        """Delete an item."""
        self.service.delete(
            id_=resource_requestctx.view_args["id"],
            identity=g.identity,
        )
        return "", 204

    @request_extra_args
    @request_view_args
    @request_headers
    @request_data
    @response_handler()
    def execute_action(self):
        """Execute action."""
        item = self.service.execute_action(
            identity=g.identity,
            id_=resource_requestctx.view_args["id"],
            action=resource_requestctx.view_args["action"],
            data=resource_requestctx.data,
            expand=resource_requestctx.args.get("expand", False),
        )
        return item.to_dict(), 200

    @request_view_args
    @request_headers
    def lock_request(self):
        """Lock a request."""
        self.service.lock_request(
            identity=g.identity,
            id_=resource_requestctx.view_args["id"],
        )
        return "", 204

    @request_view_args
    @request_headers
    def unlock_request(self):
        """Unlock a request."""
        self.service.unlock_request(
            identity=g.identity,
            id_=resource_requestctx.view_args["id"],
        )
        return "", 204



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 Graz University of Technology.
#
# Invenio-Requests is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Notification generators."""

from invenio_access.permissions import system_identity
from invenio_notifications.models import Recipient
from invenio_notifications.services.generators import RecipientGenerator
from invenio_records.dictutils import dict_lookup
from invenio_search.engine import dsl
from invenio_users_resources.proxies import current_users_service

from ..proxies import current_events_service


def _get_user_id_from_entity(entity_field):
    """Extract user ID from an entity field (expanded or non-expanded).

    :param entity_field: Entity dictionary that may contain user ID
    :returns: User ID string or None

    Examples:
        Non-expanded: {"user": "1"} -> "1"
        Expanded: {"id": "1", "profile": {...}} -> "1"
        Email entity: "user@example.com" -> None
    """
    if not isinstance(entity_field, dict):
        # e.g. resolved email entity
        return None

    # non expanded looks like {"user": "1"}
    non_expanded_id = entity_field.get("user")
    # expanded looks like {"id": "1", "profile": {"full_name": "A user"}, ... }
    # checking for None as the profile might be an empty dict (user never set a profile)
    expanded_id = (
        entity_field["id"] if entity_field.get("profile") is not None else None
    )
    return non_expanded_id or expanded_id


class RequestParticipantsRecipient(RecipientGenerator):
    """Recipient generator based on request and it's events."""

    def __init__(self, key):
        """Ctor."""
        self.key = key

    def __call__(self, notification, recipients: dict):
        """Fetch users involved in request and add as recipients."""
        request = dict_lookup(notification.context, self.key)

        # checking if entities are users. If not, we will not add them.
        # TODO: add support for other entities? (e.g. groups)
        created_by_user_id = _get_user_id_from_entity(request["created_by"])
        receiver_user_id = _get_user_id_from_entity(request["receiver"])

        user_ids = set()
        if created_by_user_id:
            user_ids.add(created_by_user_id)

        if receiver_user_id:
            user_ids.add(receiver_user_id)

        # fetching all request events to get involved users
        request_events = current_events_service.scan(
            request_id=request["id"],
            identity=system_identity,
            extra_filter=dsl.Q("term", request_id=request["id"]),
        )

        user_ids.update(
            {
                re["created_by"]["user"]
                for re in request_events
                if re["created_by"].get("user")
            }
        )

        # remove system_user_id if present
        user_ids.discard(system_identity.id)

        filter_ = dsl.Q("terms", **{"id": list(user_ids)})
        users = current_users_service.scan(system_identity, extra_filter=filter_)
        for u in users:
            recipients[u["id"]] = Recipient(data=u)
        return recipients


class CommentRepliesParticipantsRecipient(RecipientGenerator):
    """Recipient generator for comment replies thread participants.

    Collects users who are participants in a specific comment thread:
    - Parent comment author
    - All reply (child comment) authors

    This is used for reply notifications to notify only those involved
    in the specific conversation thread, not all request participants.
    """

    def __init__(self, key, event_key):
        """Ctor."""
        self.key = key
        self.event_key = event_key

    def __call__(self, notification, recipients: dict):
        """Fetch users involved in the comment thread and add as recipients."""
        request_event = dict_lookup(notification.context, self.event_key)
        request = dict_lookup(notification.context, self.key)

        # Determine the parent comment ID
        # If this event has a parent_id, use it. Otherwise, this IS the parent.
        parent_id = request_event.get("parent_id") or request_event["id"]

        user_ids = set()

        # Get parent event and its creator
        try:
            parent_event = current_events_service.read(
                identity=system_identity, id_=parent_id
            ).to_dict()

            parent_user_id = _get_user_id_from_entity(
                parent_event.get("created_by", {})
            )
            if parent_user_id:
                user_ids.add(parent_user_id)
        except Exception:
            # Parent not found or error reading, skip
            pass

        # Get all replies to the parent event for the given request
        replies_filter = dsl.Q(
            "bool",
            must=[
                dsl.Q("term", request_id=request["id"]),
                dsl.Q("term", parent_id=parent_id),
            ],
        )
        # Get all children (replies) and their creators
        # NOTE: To be improved in https://github.com/inveniosoftware/invenio-requests/issues/508
        children_events = current_events_service.scan(
            identity=system_identity,
            request_id=request["id"],
            extra_filter=replies_filter,
        )

        user_ids.update(
            {
                child["created_by"]["user"]
                for child in children_events
                if child.get("created_by", {}).get("user")
            }
        )

        # Remove system user if present
        user_ids.discard(system_identity.id)

        # Fetch users and add as recipients
        if user_ids:
            filter_ = dsl.Q("terms", **{"id": list(user_ids)})
            users = current_users_service.scan(system_identity, extra_filter=filter_)
            for u in users:
                recipients[u["id"]] = Recipient(data=u)

        return recipients



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-Requests is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.
"""User moderation service."""
from flask import current_app
from invenio_accounts.models import Role
from invenio_i18n import gettext as _
from invenio_records_resources.services.uow import unit_of_work
from invenio_search.engine import dsl

from invenio_requests.customizations.user_moderation.user_moderation import (
    UserModerationRequest,
)
from invenio_requests.proxies import current_request_type_registry
from invenio_requests.services.user_moderation.errors import OpenRequestAlreadyExists

from ..results import EntityResolverExpandableField


class UserModerationRequestService:
    """Service for User Moderation requests."""

    def __init__(self, requests_service):
        """Service initialisation as a sub-service of requests."""
        self.requests_service = requests_service

    @property
    def request_type_cls(self):
        """User moderation request type."""
        return current_request_type_registry.lookup(UserModerationRequest.type_id)

    def _exists(self, identity, user_id):
        """Return the request id if an open request already exists, else None."""
        results = self.requests_service.search(
            identity,
            extra_filter=dsl.query.Bool(
                "must",
                must=[
                    dsl.Q("term", **{"type": self.request_type_cls.type_id}),
                    dsl.Q("term", **{"topic.user": user_id}),
                    dsl.Q("term", **{"is_open": True}),
                ],
            ),
        )
        return next(results.hits)["id"] if results.total > 0 else None

    def _create_request(self, identity, user_id, creator, receiver, data, uow):
        """Creates and submits the request."""
        if self._exists(identity, user_id):
            raise OpenRequestAlreadyExists

        # For user moderation, topic is the user to be moderated
        topic = {"user": str(user_id)}

        # Receiver can be configured, by default send the request to users with moderation role
        receiver = {"group": receiver}  # TODO to be changed to role id
        creator = {"group": creator}  # TODO to be changed to role id

        request_item = self.requests_service.create(
            identity,
            data,
            self.request_type_cls,
            receiver,
            creator,
            topic=topic,
            uow=uow,
        )

        # Permission for identity might be denied on 'submit'
        return self.requests_service.execute_action(
            identity=identity, id_=request_item.id, action="submit", data=data, uow=uow
        )

    @unit_of_work()
    def request_moderation(self, identity, user_id, data=None, uow=None, **kwargs):
        """Creates a UserModeration request and submits it."""
        REQUESTS_MODERATION_ROLE = current_app.config.get("REQUESTS_MODERATION_ROLE")
        role = Role.query.filter(Role.name == REQUESTS_MODERATION_ROLE).one_or_none()
        assert role, _("Moderation role must exist to enable user moderation requests.")

        data = data or {}

        return self._create_request(
            identity,
            user_id,
            creator=role.name,
            receiver=role.name,
            data=data,
            uow=uow,
        )

    @property
    def expandable_fields(self):
        """Get expandable fields."""
        return [
            EntityResolverExpandableField("created_by"),
            EntityResolverExpandableField("receiver"),
            EntityResolverExpandableField("topic"),
        ]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
# Copyright (C) 2021 Northwestern University.
# Copyright (C) 2021 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Request Events Service Config."""

from invenio_indexer.api import RecordIndexer
from invenio_records_resources.services import (
    Link,
    RecordServiceConfig,
    ServiceSchemaWrapper,
)
from invenio_records_resources.services.base.config import ConfiguratorMixin, FromConfig
from invenio_records_resources.services.records.links import pagination_links
from invenio_records_resources.services.records.results import (
    RecordItem,
    RecordList,
)

from invenio_requests.proxies import (
    current_request_type_registry,
)

from ...records.api import Request, RequestEvent
from ..permissions import PermissionPolicy
from ..schemas import RequestEventSchema


class RequestEventItem(RecordItem):
    """RequestEvent result item."""

    def __init__(self, *args, **kwargs):
        """Constructor."""
        request = kwargs.pop("request", None)
        super().__init__(*args, **kwargs)
        self._request = request

    @property
    def id(self):
        """Id property."""
        return self._record.id

    @property
    def data(self):
        """Property to get the record projection with request in context."""
        if self._data:
            return self._data

        self._data = self._schema.dump(
            self._obj,
            context=dict(
                identity=self._identity,
                record=self._record,
                request=self._request,  # Need to pass the request to the schema to get the permissions to check if locked
            ),
        )
        if self._links_tpl:
            self._data["links"] = self.links

        if self._nested_links_item:
            for link in self._nested_links_item:
                link.expand(self._identity, self._record, self._data)

        if self._expand and self._fields_resolver:
            self._fields_resolver.resolve(self._identity, [self._data])
            fields = self._fields_resolver.expand(self._identity, self._data)
            self._data["expanded"] = fields

        return self._data


class RequestEventList(RecordList):
    """RequestEvent result item."""

    def __init__(self, *args, **kwargs):
        """Constructor."""
        request = kwargs.pop("request", None)
        super().__init__(*args, **kwargs)
        self._request = request

    def to_dict(self):
        """Return result as a dictionary with expanded fields for parents and children."""
        # Call parent to handle standard expansion
        res = super().to_dict()

        # Additionally expand children fields if present
        if self._expand and self._fields_resolver:
            self._expand_children_fields(res["hits"]["hits"])

        return res

    def _expand_children_fields(self, hits):
        """Apply field expansion to children arrays in hits.

        :param hits: List of hit dictionaries that may contain children arrays
        """
        # Collect all children from all hits
        all_children = []
        for hit in hits:
            if "children" in hit and hit["children"]:
                all_children.extend(hit["children"])

        if all_children:
            # Batch resolve all children at once for efficiency
            self._fields_resolver.resolve(self._identity, all_children)

            # Expand each child individually
            for child in all_children:
                fields = self._fields_resolver.expand(self._identity, child)
                child["expanded"] = fields

    @property
    def hits(self):
        """Iterator over the hits."""
        for hit in self._results:
            # Load dump
            record = self._service.record_cls.loads(hit.to_dict())

            # Project the record
            schema = ServiceSchemaWrapper(
                self._service, record.type.marshmallow_schema()
            )
            projection = schema.dump(
                record,
                context=dict(
                    identity=self._identity,
                    record=record,
                    request=self._request,  # Need to pass the request to the schema to get the permissions to check if locked
                    meta=hit.meta,
                ),
            )

            # Handle inner_hits from has_child queries (join relationship approach)
            # Initialize defaults for parents without children
            projection["children"] = []
            projection["children_count"] = 0

            if (
                hasattr(hit.meta, "inner_hits")
                and "replies_preview" in hit.meta.inner_hits
            ):
                # Extract children from inner_hits
                inner_hits_data = hit.meta.inner_hits.replies_preview.hits
                inner_children = inner_hits_data.hits
                total_children = inner_hits_data.total.value

                projection["children_count"] = total_children

                for inner_hit in inner_children:
                    # Load child record
                    child_record = self._service.record_cls.loads(
                        inner_hit["_source"].to_dict()
                    )

                    # Project child record
                    child_schema = ServiceSchemaWrapper(
                        self._service, child_record.type.marshmallow_schema()
                    )
                    child_projection = child_schema.dump(
                        child_record,
                        context=dict(
                            identity=self._identity,
                            record=child_record,
                            request=self._request,  # Need to pass the request to the schema to get the permissions to check if locked
                            meta=hit.meta,
                        ),
                    )

                    if self._links_item_tpl:
                        child_projection["links"] = self._links_item_tpl.expand(
                            self._identity, child_record
                        )

                    projection["children"].append(child_projection)

            if self._links_item_tpl:
                projection["links"] = self._links_item_tpl.expand(
                    self._identity, record
                )

            yield projection


class RequestEventLink(Link):
    """Link variables setter for RequestEvent links."""

    @staticmethod
    def vars(obj, vars):
        """Variables for the URI template."""
        request_type = current_request_type_registry.lookup(vars["request_type"])
        parent_id = obj.parent_id if obj.parent_id else obj.id
        vars.update(
            {"id": obj.id, "request_id": obj.request_id, "parent_id": parent_id}
        )
        vars.update(request_type._update_link_config(**vars))


class ParentChildRecordIndexer(RecordIndexer):
    """Parent-Child Record Indexer placeholder."""

    def _prepare_record(self, record, index, arguments=None, **kwargs):
        """Prepare request-event data for indexing.

        Pass routing information for parent-child relationships.
        """
        data = super()._prepare_record(record, index, arguments, **kwargs)
        if hasattr(record, "parent_id") and record.parent_id:
            arguments["routing"] = str(record.parent_id)
        return data


class RequestEventsServiceConfig(RecordServiceConfig, ConfiguratorMixin):
    """Config."""

    service_id = "request_events"

    request_cls = Request
    permission_policy_cls = FromConfig(
        "REQUESTS_PERMISSION_POLICY", default=PermissionPolicy
    )
    schema = RequestEventSchema
    record_cls = RequestEvent
    result_item_cls = RequestEventItem
    result_list_cls = RequestEventList
    indexer_queue_name = "events"
    indexer_cls = ParentChildRecordIndexer

    # ResultItem configurations
    links_item = {
        "self": RequestEventLink("{+api}/requests/{request_id}/comments/{id}"),
        "self_html": RequestEventLink("{+ui}/requests/{request_id}#commentevent-{id}"),
        "reply": RequestEventLink(
            "{+api}/requests/{request_id}/comments/{parent_id}/reply"
        ),
        "replies": RequestEventLink(
            "{+api}/requests/{request_id}/comments/{parent_id}/replies"
        ),
    }
    links_search = pagination_links("{+api}/requests/{request_id}/timeline{?args*}")

    links_replies = pagination_links(
        "{+api}/requests/{request_id}/comments/{parent_id}/replies{?args*}"
    )

    components = FromConfig(
        "REQUESTS_EVENTS_SERVICE_COMPONENTS",
        default=[],
    )
